import requests
import unittest


@unittest.skip
class RequestTest(unittest.TestCase):
    def test_get_returns_200(self):
        r = requests.get("https://jsonplaceholder.typicode.com/todos/1")
        self.assertEqual(r.status_code, 200)

    def test_get_with_params_returns_json(self):
        payload = {"postId": 1}
        r = requests.get(
            "https://jsonplaceholder.typicode.com/comments", params=payload
        )
        self.assertEqual(len(r.json()), 5)

    def test_get_with_params_returns_json_with_expected_attributes(self):
        expected_post_id = 1
        payload = {"postId": expected_post_id}
        r = requests.get("http://jsonplaceholder.typicode.com/comments", params=payload)
        json = r.json()

        for index in range(len(json)):
            json_item = json[index]
            actual_id = json_item["postId"]
            self.assertEquals(actual_id, expected_post_id)

    def test_post_json(self):
        payload = {
            "title": "tttttiiihihihitle",
            "body": "aaaaaaaahahaha",
            "userId": 1000,
        }

        response = requests.post(
            "http://jsonplaceholder.typicode.com/posts", json=payload
        )

        self.assertEqual(response.status_code, 201)



import copy
import logging
from lastipy.lastfm.library.recent_artists import fetch_recent_artists


def calculate_ratings(
    user, api_key, top_tracks_to_recommendations, prefer_unheard_artists=True
):
    """Returns a copy of the list of recommendations in the given map, with ratings set based on recommendation
    'strength', which can be used to determine its chances of showing up in the final playlist"""

    logging.info("Adjusting track recommendation ratings")

    # Create a copy of the given map and modify it instead
    top_tracks_to_recommendations_copy = copy.deepcopy(top_tracks_to_recommendations)

    _adjust_ratings_based_on_playcounts(top_tracks_to_recommendations_copy)

    if prefer_unheard_artists:
        _adjust_ratings_based_on_recent_artists(
            top_tracks_to_recommendations_copy, user, api_key
        )

    logging.info("Finished adjusting ratings")
    return _extract_tracks_from_map(top_tracks_to_recommendations_copy)


def _adjust_ratings_based_on_playcounts(top_tracks_to_recommendations):
    """Adjust rating based on associated top track's playcount; the thought being that recommendations based on more listened-to
    tracks should have a higher chance of being recommended"""
    for top_track in top_tracks_to_recommendations:
        recommendations = top_tracks_to_recommendations[top_track]
        for recommendation in recommendations:
            recommendation.recommendation_rating *= top_track.playcount


def _adjust_ratings_based_on_recent_artists(
    top_tracks_to_recommendations, user, api_key
):
    """Reduces tracks ratings based on their artists' playcount in the user's library"""
    recent_artists = fetch_recent_artists(user, api_key)
    for top_track in top_tracks_to_recommendations:
        recommendations = top_tracks_to_recommendations[top_track]
        for recommendation in recommendations:
            for artist in recent_artists:
                if recommendation.artist.lower() == artist.artist_name.lower():
                    # Reduce the rating by dividing it by the artist's playcount; the thought being that the more listened-to an artist is,
                    # the less chance there should be of one of their tracks being recommended
                    recommendation.recommendation_rating /= artist.playcount
                    logging.debug(
                        "Calculated rating for "
                        + str(recommendation)
                        + ": "
                        + str(recommendation.recommendation_rating)
                    )


def _extract_tracks_from_map(top_tracks_to_recommendations):
    all_recommendations = []
    for top_track in top_tracks_to_recommendations:
        recommendations = top_tracks_to_recommendations[top_track]
        all_recommendations = all_recommendations + recommendations
    return all_recommendations



#!/usr/bin/env python3.7

from configparser import ConfigParser
import argparse
import os
from lastipy import definitions
from lastipy.lastfm.library.top_tracks import fetch_top_tracks
from lastipy.lastfm.library.recent_tracks import fetch_recent_tracks
from lastipy.lastfm.library.recent_artists import fetch_recent_artists
from lastipy.util.filter import filter_out_tracks_in_second_list
from lastipy.lastfm.library import period
from lastipy.spotify import playlist, search, library
from lastipy.track import Track
from numpy.random import choice
from spotipy import Spotify
from lastipy.spotify import token
from lastipy.util.setup_logging import setup_logging
import logging
from lastipy.spotify import new_releases
from datetime import datetime
from lastipy.util.parse_api_keys import ApiKeysParser
from lastipy.spotify import album
from datetime import date, timedelta


# TODO test
def save_new_releases():
    """Saves all new releases released yesterday by the specified Spotify user's followed artists to their library. We check
    for yesterday so that all releases released throughout that day are picked up (trying to get today's releases has a chance of missing anything
    released after this script is run)."""

    setup_logging("new_releases.log")

    args = _extract_args()

    spotify = Spotify(
        auth=token.get_token(
            args.spotify_user,
            args.spotify_client_id_key,
            args.spotify_client_secret_key,
        )
    )

    if args.save_albums_to_liked_songs:
        _save_new_tracks(
            spotify,
            args.lastfm_user,
            args.lastfm_api_key,
            args.ignore_remixes,
            args.ignore_scrobbled_songs,
            # Save all types of albums straight to Liked Songs
            [album.SINGLE_ALBUM_TYPE, album.ALBUM_ALBUM_TYPE],
        )
    else:
        _save_new_tracks(
            spotify,
            args.lastfm_user,
            args.lastfm_api_key,
            args.ignore_remixes,
            args.ignore_scrobbled_songs,
            # Save only single-type albums to Liked Songs
            [album.SINGLE_ALBUM_TYPE],
        )

        new_albums = new_releases.fetch_new_albums(
            spotify, album_types=[album.ALBUM_ALBUM_TYPE], as_of_date=yesterday
        )
        if len(new_albums) > 0:
            library.add_albums_to_library(spotify, new_albums)
        else:
            logging.info("No new albums to add!")

    logging.info("Done!")


def _save_new_tracks(
    spotify,
    lastfm_user,
    lastfm_api_key,
    ignore_remixes,
    ignore_scrobbled_songs,
    album_types,
):
    yesterday = date.today() - timedelta(days=1)

    new_tracks = new_releases.fetch_new_tracks(
        spotify,
        ignore_remixes=ignore_remixes,
        album_types=album_types,
        as_of_date=yesterday,
    )

    if ignore_scrobbled_songs:
        logging.info("Filtering out scrobbled tracks from new releases")
        # TODO this is pretty inefficient but it appears to be the only way to tell if a user has scrobbled a track
        recent_tracks = fetch_recent_tracks(user=lastfm_user, api_key=lastfm_api_key)
        new_tracks = filter_out_tracks_in_second_list(new_tracks, recent_tracks)

    if len(new_tracks) > 0:
        library.add_tracks_to_library(spotify, new_tracks)
    else:
        logging.info("No tracks to add!")


def _extract_args():
    args = _parse_args()
    _extract_api_keys(args)
    _extract_user_configs(args)
    return args


def _extract_api_keys(args):
    keys_parser = ApiKeysParser(args.api_keys_file)
    args.spotify_client_id_key = keys_parser.spotify_client_id_key
    args.spotify_client_secret_key = keys_parser.spotify_client_secret_key
    args.lastfm_api_key = keys_parser.lastfm_api_key


def _extract_user_configs(args):
    config_parser = ConfigParser()
    config_parser.read(args.user_configs_file.name)
    section = "Config"
    args.spotify_user = config_parser.get(section, "SpotifyUser")
    args.lastfm_user = config_parser.get(section, "LastFMUser")
    args.ignore_remixes = config_parser.getboolean(section, "IgnoreRemixes")
    args.ignore_scrobbled_songs = config_parser.getboolean(
        section, "IgnoreScrobbledSongs"
    )
    args.save_albums_to_liked_songs = config_parser.getboolean(
        section, "SaveAlbumsToLikedSongs"
    )
    return args


def _parse_args():
    args_parser = argparse.ArgumentParser(
        description="Adds new tracks from the given user's followed artists to their saved/liked tracks"
    )
    args_parser.add_argument(
        "user_configs_file", type=argparse.FileType("r", encoding="UTF-8")
    )
    args_parser.add_argument(
        "api_keys_file", type=argparse.FileType("r", encoding="UTF-8")
    )
    return args_parser.parse_args()


if __name__ == "__main__":
    save_new_releases()



import unittest
from unittest.mock import patch
from lastipy.lastfm.library import recent_artists
from unittest.mock import Mock
from requests import HTTPError
from lastipy.lastfm.library.scrobbled_artist import ScrobbledArtist


class RecentArtistsTest(unittest.TestCase):

    dummy_user = "dummyUser"
    dummy_api_key = "123456789"

    @patch("lastipy.lastfm.library.recent_artists.fetch_paginated_response")
    def test_fetch(self, mock_paginated_endpoint_fetch):
        expected_artists = [
            ScrobbledArtist(artist_name="The Beatles", playcount=10),
            ScrobbledArtist(artist_name="Cream", playcount=4),
            ScrobbledArtist(artist_name="Led Zeppelin", playcount=7),
        ]

        mock_paginated_endpoint_fetch.return_value = [
            {
                "artists": {
                    "artist": [
                        {"name": "The Beatles", "playcount": "10"},
                        {"name": "Cream", "playcount": "4"},
                    ],
                    "@attr": {"totalPages": "2"},
                }
            },
            {
                "artists": {
                    "artist": [{"name": "Led Zeppelin", "playcount": "7"}],
                    "@attr": {"totalPages": "2"},
                }
            },
        ]

        fetched_artists = recent_artists.fetch_recent_artists(
            user=self.dummy_user, api_key=self.dummy_api_key
        )

        self.assertCountEqual(fetched_artists, expected_artists)



import unittest
from lastipy.spotify.library import add_albums_to_library
from spotipy import Spotify
from lastipy.spotify.album import SpotifyAlbum


class AddAlbumsToLibraryTest(unittest.TestCase):
    def test_adding_less_than_max_request(self):
        spotify = Spotify()
        spotify.current_user = unittest.mock.MagicMock({"id": "testUser"})
        spotify.current_user_saved_albums_add = unittest.mock.MagicMock()
        dummy_albums = [
            self._build_dummy_album("123"),
            self._build_dummy_album("456"),
            self._build_dummy_album("789"),
        ]
        add_albums_to_library(spotify, dummy_albums)
        spotify.current_user_saved_albums_add.assert_called_with(["123", "456", "789"])

    def test_adding_more_than_max_request(self):
        spotify = Spotify()
        spotify.current_user = unittest.mock.MagicMock({"id": "testUser"})
        spotify.current_user_saved_albums_add = unittest.mock.MagicMock()

        dummy_albums = []
        for _ in range(150):
            dummy_albums.append(self._build_dummy_album("123"))

        expected_chunks = []
        for _ in range(3):
            chunk = []
            for _ in range(50):
                chunk.append("123")
            expected_chunks.append(chunk)

        add_albums_to_library(spotify, dummy_albums)

        for i in range(3):
            spotify.current_user_saved_albums_add.assert_called_with(expected_chunks[i])

    def _build_dummy_album(self, spotify_id):
        return SpotifyAlbum(
            name="test_name",
            artist="test_artist",
            spotify_id=spotify_id,
            album_type="album",
            release_date_precision="day",
            release_date="2000-01-01",
        )



class SpotifyAlbum:
    """Represents a Spotify "album". This is slightly confusing in that there are different "types" of albums: albums, and singles."""

    def __init__(
        self, name, artist, album_type, spotify_id, release_date_precision, release_date
    ):
        self.name = name
        self.artist = artist
        self.album_type = album_type
        self.spotify_id = spotify_id
        self.release_date_precision = release_date_precision
        self.release_date = release_date

    def __eq__(self, other):
        return (
            isinstance(other, SpotifyAlbum)
            and self.name == other.name
            and self.artist == other.artist
            and self.album_type == other.album_type
            and self.spotify_id == other.spotify_id
            and self.release_date_precision == self.release_date_precision
            and self.release_date == self.release_date
        )

    def __repr__(self):
        return str(self.__dict__)


SINGLE_ALBUM_TYPE = "single"
ALBUM_ALBUM_TYPE = "album"



import unittest
from unittest.mock import MagicMock
from spotipy import Spotify
from lastipy.spotify.library import get_saved_tracks
from lastipy.track import Track


class GetSavedTracksTest(unittest.TestCase):
    def setUp(self):
        self.mock_spotify = Spotify()
        self.mock_spotify.current_user = MagicMock({"id": "dummyUser"})

    def test_fetch_single_page(self):
        expected_saved_tracks = [
            Track(track_name="Penny Lane", artist="The Beatles", spotify_id="123456789")
        ]

        self.mock_spotify.current_user_saved_tracks = MagicMock()
        mock_saved_tracks_response = {
            "items": [
                {
                    "id": "123456789",
                    "name": "Penny Lane",
                    "artists": [{"name": "The Beatles"}],
                }
            ]
        }
        self.mock_spotify.current_user_saved_tracks.side_effect = [
            mock_saved_tracks_response,
            {"items": []},
        ]

        fetched_saved_tracks = get_saved_tracks(self.mock_spotify)

        self.assertCountEqual(fetched_saved_tracks, expected_saved_tracks)

    def test_fetch_multiple_pages(self):
        expected_saved_tracks = [
            Track(
                track_name="Penny Lane", artist="The Beatles", spotify_id="123456789"
            ),
            Track(
                track_name="A Day in the Life",
                artist="The Beatles",
                spotify_id="987654321",
            ),
        ]

        self.mock_spotify.current_user_saved_tracks = MagicMock()
        mock_saved_tracks_response_page_1 = {
            "items": [
                {
                    "id": "123456789",
                    "name": "Penny Lane",
                    "artists": [{"name": "The Beatles"}],
                }
            ]
        }
        mock_saved_tracks_response_page_2 = {
            "items": [
                {
                    "id": "987654321",
                    "name": "A Day in the Life",
                    "artists": [{"name": "The Beatles"}],
                }
            ]
        }
        self.mock_spotify.current_user_saved_tracks.side_effect = [
            mock_saved_tracks_response_page_1,
            mock_saved_tracks_response_page_2,
            {"items": []},
        ]

        fetched_saved_tracks = get_saved_tracks(self.mock_spotify)

        self.assertCountEqual(fetched_saved_tracks, expected_saved_tracks)



import logging
import requests
from lastipy.recommendations.recommended_track import RecommendedTrack
from lastipy.lastfm.parse_lastfm_tracks import parse_track_name, parse_artist

URL = "http://ws.audioscrobbler.com/2.0/?method=track.getsimilar"


def fetch_recommendations(api_key, track, limit):
    """Fetches tracks similar to the given track.
    :param limit: the maximum number of items to fetch"""

    logging.info(
        "Fetching up to "
        + str(limit)
        + " recommendations based on "
        + str(track)
        + " in Last.fm"
    )
    json_response = _send_request(_build_json_payload(api_key, track, limit))
    if "similartracks" in json_response:
        json_tracks = json_response["similartracks"]["track"]
        similar_tracks = [
            RecommendedTrack(
                track_name=parse_track_name(json_track), artist=parse_artist(json_track)
            )
            for json_track in json_tracks
        ]
        logging.info("Fetched " + str(len(similar_tracks)) + " track recommendations")
        logging.debug("Fetched tracks: " + str(similar_tracks))
        return similar_tracks
    elif "errors" in json_response:
        raise Exception(
            "Error occurred while fetching similar tracks " + json_response["errors"]
        )


def _send_request(json_payload):
    response = requests.get(URL, params=json_payload)
    if response.ok:
        return response.json()
    else:
        response.raise_for_status()


def _build_json_payload(api_key, track, limit):
    payload = {
        "track": track.track_name,
        "artist": track.artist,
        "format": "json",
        "api_key": api_key,
        "limit": limit,
    }
    return payload



import logging
from lastipy.spotify.search import search_for_tracks
from lastipy.recommendations.recommended_track import RecommendedTrack

# TODO test
def fetch_recommendations(spotify, track, limit):
    logging.info(
        "Fetching up to "
        + str(limit)
        + " recommendations based on "
        + str(track)
        + " in Spotify"
    )
    track_in_spotify = search_for_tracks(
        spotify=spotify, query=track.artist + " " + track.track_name
    )[0]
    # TODO this accepts up to 5 tracks, so this could be more sophisticated by using multiple top-listened-to tracks in a single call
    recommendations = spotify.recommendations(
        seed_tracks=[track_in_spotify.spotify_id], limit=limit
    )

    parsed_recommendations = []
    for recommendation in recommendations["tracks"]:
        parsed_recommendations.append(
            RecommendedTrack(
                track_name=recommendation["name"],
                artist=recommendation["artists"][0]["name"],
                spotify_id=recommendation["id"],
                recommendation_rating=1,
            )
        )

    logging.info(
        "Fetched " + str(len(parsed_recommendations)) + " track recommendations"
    )
    return parsed_recommendations



def chunk_list(list_to_be_chunked, chunk_size):
    """Splits the given list into evenly-sized "chunks", each of size chunk_size"""
    return [
        list_to_be_chunked[i : i + chunk_size]
        for i in range(0, len(list_to_be_chunked), chunk_size)
    ]



#!/usr/bin/env python3.7

from configparser import ConfigParser
import argparse
import os
from lastipy import definitions
from lastipy.lastfm.library import period
from lastipy.track import Track
from lastipy.util.setup_logging import setup_logging
import logging
from lastipy.util.parse_api_keys import ApiKeysParser
from spotipy import Spotify
from lastipy.spotify import token
from lastipy.recommendations.recommendations import generate_recommendations
from lastipy.spotify import playlist, search
from numpy.random import choice


# TODO test
def build_recommendations_playlist():
    """Adds recommendations to the given Spotify user's playlist based on the given Last.fm user's recommendations"""

    setup_logging("recommendations.log")
    args = _extract_args()

    spotify = Spotify(
        auth=token.get_token(
            args.spotify_user,
            args.spotify_client_id_key,
            args.spotify_client_secret_key,
        )
    )

    recommendations = generate_recommendations(
        lastfm_user=args.lastfm_user,
        lastfm_api_key=args.lastfm_api_key,
        spotify=spotify,
        recommendation_services=args.recommendation_services,
        recommendation_period=args.recommendation_period,
        max_recommendations_per_top_track=args.max_recommendations_per_top_track,
        blacklisted_artists=args.blacklisted_artists,
        prefer_unheard_artists=args.prefer_unheard_artists,
    )

    logging.info("Generating recommendations playlist...")
    tracks_for_playlist = []
    while len(tracks_for_playlist) < args.playlist_size and len(recommendations) > 0:
        # Find and remove the recommendation from the list using a random choice, weighted based on the recommendation's rating value
        recommendation = choice(
            recommendations, p=_calculate_rating_weights(recommendations)
        )
        recommendations.remove(recommendation)

        if recommendation.spotify_id is None:
            # If we didn't get the recommendation directly from Spotify, we need to get the ID from there before we can add to the playlist
            search_results = search.search_for_tracks(
                spotify=spotify,
                query=recommendation.artist + " " + recommendation.track_name,
            )
            if search_results and Track.are_equivalent(
                recommendation, search_results[0]
            ):
                recommendation.spotify_id = search_results[0].spotify_id

        if recommendation.spotify_id is not None and not any(
            recommendation.artist == item.artist for item in tracks_for_playlist
        ):
            logging.info("Adding " + str(recommendation))
            tracks_for_playlist.append(recommendation)

    playlist.replace_tracks_in_playlist(
        spotify, args.playlist_name, tracks_for_playlist
    )

    logging.info("Done!")


def _calculate_rating_weights(recommendations):
    logging.debug("Calculating recommendation rating weights")
    total_ratings = 0
    for recommendation in recommendations:
        total_ratings += recommendation.recommendation_rating

    rating_weights = []
    for recommendation in recommendations:
        rating_weights.append(recommendation.recommendation_rating / total_ratings)
    return rating_weights


def _extract_args():
    args_parser = _setup_args_parser()
    args = args_parser.parse_args()
    args = _extract_user_configs(args)
    args = _extract_api_keys(args)
    return args


def _setup_args_parser():
    parser = argparse.ArgumentParser(
        description="Create a Spotify playlist based off recommendations from Last.fm"
    )
    parser.add_argument(
        "user_configs_file", type=argparse.FileType("r", encoding="UTF-8")
    )
    parser.add_argument("api_keys_file", type=argparse.FileType("r", encoding="UTF-8"))
    return parser


def _extract_user_configs(args):
    config_parser = ConfigParser()
    config_parser.read(args.user_configs_file.name)
    section = "Config"
    args.lastfm_user = config_parser[section]["LastFMUser"]
    args.spotify_user = config_parser[section]["SpotifyUser"]
    args.recommendation_services = config_parser[section][
        "RecommendationServices"
    ].split(",")
    args.recommendation_period = config_parser[section]["RecommendationPeriod"]
    args.max_recommendations_per_top_track = int(
        config_parser[section]["MaxRecommendationsPerTopTrack"]
    )
    args.playlist_size = int(config_parser[section]["PlaylistSize"])
    args.playlist_name = config_parser[section]["PlaylistName"]
    args.blacklisted_artists = config_parser[section]["BlacklistedArtists"].split(",")
    args.prefer_unheard_artists = _str_to_bool(
        config_parser[section]["PreferUnheardArtists"]
    )
    return args


def _str_to_bool(to_convert):
    return to_convert == "True"


def _extract_api_keys(args):
    keys_parser = ApiKeysParser(args.api_keys_file)
    args.lastfm_api_key = keys_parser.lastfm_api_key
    args.spotify_client_id_key = keys_parser.spotify_client_id_key
    args.spotify_client_secret_key = keys_parser.spotify_client_secret_key
    return args


if __name__ == "__main__":
    build_recommendations_playlist()



import logging
from lastipy.lastfm.library import period
from lastipy.lastfm.parse_lastfm_tracks import parse_track_name, parse_artist
from lastipy.lastfm.library.top_track import TopTrack
from lastipy.lastfm.library.paginated_endpoint import fetch_paginated_response

URL = "http://ws.audioscrobbler.com/2.0/?method=user.gettoptracks"


def fetch_top_tracks(user, api_key, a_period=period.OVERALL):
    """Fetches the top tracks for the given user over the given period"""

    logging.info("Fetching top tracks for user " + user + " over period " + a_period)

    paginated_json_responses = fetch_paginated_response(
        URL, user, api_key, "toptracks", [{"key": "period", "value": a_period}]
    )

    top_tracks = []
    for json_response in paginated_json_responses:
        for track in json_response["toptracks"]["track"]:
            top_tracks.append(
                TopTrack(
                    parse_track_name(track),
                    parse_artist(track),
                    int(track["playcount"]),
                )
            )

    # Filter out tracks with a playcount of 1, since those really shouldn't be considered "top"
    top_tracks = [track for track in top_tracks if track.playcount > 1]

    logging.info("Fetched " + str(len(top_tracks)) + " top tracks")
    logging.debug("Fetched tracks: " + str(top_tracks))
    return top_tracks



import spotipy.oauth2 as oauth2
import webbrowser
import os
from lastipy import definitions

REDIRECT_URI = "https://www.example.com/callback/"


# TODO test
def get_token(username, client_id_key, client_secret_key):
    """Returns a Spotify token for the given user. If a cached token file exists (with the format .cache-<username>),
    it is returned; otherwise, the given user will be prompted to authorize the app.
    This function was copied from util.py in spotipy and modified in order to customize the path to the cached token."""

    # These are the only scopes required by this app so no need to parameterize this
    scope = (
        "playlist-modify-public user-library-read user-library-modify user-follow-read"
    )

    sp_oauth = oauth2.SpotifyOAuth(
        client_id_key,
        client_secret_key,
        REDIRECT_URI,
        scope=scope,
        cache_path=os.path.join(definitions.ROOT_DIR, ".cache-" + username),
    )

    token_info = sp_oauth.get_cached_token()

    if not token_info:
        print(
            """
            User authentication requires interaction with your
            web browser. Once you enter your credentials and
            give authorization, you will be redirected to
            a url.  Paste that url you were directed to to
            complete the authorization.
        """
        )
        auth_url = sp_oauth.get_authorize_url()
        try:
            webbrowser.open(auth_url)
            print("Opened %s in your browser" % auth_url)
        except:
            print("Please navigate here: %s" % auth_url)

        print()
        print()
        response = input("Enter the URL you were redirected to: ")

        print()
        print()

        code = sp_oauth.parse_response_code(response)
        token_info = sp_oauth.get_access_token(code)

    if token_info:
        return token_info["access_token"]
    else:
        return None



import unittest
from unittest.mock import patch
from unittest.mock import Mock
from requests import HTTPError
from lastipy.track import Track
from lastipy.lastfm.library.paginated_endpoint import fetch_paginated_response


class RecentArtistsTest(unittest.TestCase):

    dummy_url = "www.example.org"
    dummy_user = "dummyUser"
    dummy_api_key = "123456789"

    @patch("requests.get")
    def test_fetch_single_page(self, mock_requests_get):
        mock_response = Mock()
        mock_response.ok = True
        json_object_page = {
            "arrayKey": {
                "array": [
                    {"key1": "Value1", "key2": {"key3": "Value2"}},
                    {"key1": "Value3", "key2": {"key3": "Value4"}},
                ],
                "@attr": {"totalPages": "1"},
            }
        }
        mock_response.json.return_value = json_object_page
        mock_requests_get.side_effect = [mock_response]

        fetched_objects = fetch_paginated_response(
            self.dummy_url, self.dummy_user, self.dummy_api_key, "arrayKey"
        )

        self.assertCountEqual(fetched_objects, [json_object_page])

    @patch("requests.get")
    def test_fetch_multiple_pages(self, mock_requests_get):
        json_object_page_1 = {
            "arrayKey": {
                "array": [
                    {"key1": "Value1", "key2": "Value2"},
                    {"key1": "Value3", "key2": "Value4"},
                ],
                "@attr": {"totalPages": "3"},
            }
        }
        json_object_page_2 = {
            "arrayKey": {
                "array": [
                    {"key1": "Value5", "key2": "Value6"},
                    {"key1": "Value7", "key2": "Value8"},
                ],
                "@attr": {"totalPages": "3"},
            }
        }
        json_object_page_3 = {
            "arrayKey": {
                "array": [
                    {"key1": "Value9", "key2": "Value10"},
                    {"key1": "Value11", "key2": "Value12"},
                ],
                "@attr": {"totalPages": "3"},
            }
        }

        mock_responses = [Mock(), Mock(), Mock()]
        mock_responses[0].ok = True
        mock_responses[0].json.return_value = json_object_page_1
        mock_responses[1].ok = True
        mock_responses[1].json.return_value = json_object_page_2
        mock_responses[2].ok = True
        mock_responses[2].json.return_value = json_object_page_3
        mock_requests_get.side_effect = mock_responses

        fetched_objects = fetch_paginated_response(
            self.dummy_url, self.dummy_user, self.dummy_api_key, "arrayKey"
        )

        self.assertCountEqual(
            fetched_objects,
            [json_object_page_1, json_object_page_2, json_object_page_3],
        )

    @patch("requests.get")
    def test_fetch_with_success_after_retries(self, mock_requests_get):
        json_object_page_1 = {
            "arrayKey": {
                "array": [
                    {"key1": "Value1", "key2": "Value2"},
                    {"key1": "Value3", "key2": "Value4"},
                ],
                "@attr": {"totalPages": "2"},
            }
        }

        json_object_page_2 = {
            "arrayKey": {
                "array": [
                    {"key1": "Value5", "key2": "Value6"},
                    {"key1": "Value7", "key2": "Value8"},
                ],
                "@attr": {"totalPages": "2"},
            }
        }

        mock_responses = [Mock(), Mock(), Mock()]
        mock_responses[0].ok = True
        mock_responses[0].json.return_value = json_object_page_1
        mock_responses[1].ok = False
        mock_responses[1].raise_for_status.side_effect = HTTPError(
            Mock(status=500), "Error"
        )
        mock_responses[2].ok = True
        mock_responses[2].json.return_value = json_object_page_2
        mock_requests_get.side_effect = mock_responses

        fetched_objects = fetch_paginated_response(
            self.dummy_url, self.dummy_user, self.dummy_api_key, "arrayKey"
        )

        self.assertCountEqual(fetched_objects, [json_object_page_1, json_object_page_2])

    @patch("requests.get")
    def test_fetch_fails_after_retries(self, mock_requests_get):
        mock_responses = []
        for _ in range(11):
            mock_response = Mock()
            mock_response.ok = False
            mock_response.raise_for_status.side_effect = HTTPError(
                Mock(status=500), "Error"
            )
            mock_responses.append(mock_response)

        # Add another mock response, but the code will exit after the retry limit is reached and this won't actually get fetched
        mock_response = Mock()
        mock_response.ok = True
        mock_response.json.return_value = {"test"}
        mock_responses.append(mock_response)

        mock_requests_get.side_effect = mock_responses

        fetched_objects = fetch_paginated_response(
            self.dummy_url, self.dummy_user, self.dummy_api_key, "arrayKey"
        )

        self.assertCountEqual(fetched_objects, [])

    @patch("requests.get")
    def test_extra_request_params(self, mock_requests_get):
        mock_requests_get.ok = True
        mock_requests_get.json.return_value = []

        fetch_paginated_response(
            self.dummy_url,
            self.dummy_user,
            self.dummy_api_key,
            "arrayKey",
            [{"key": "key1", "value": "value1"}, {"key": "key2", "value": "value2"}],
        )

        expected_request = {
            "user": self.dummy_user,
            "format": "json",
            "api_key": self.dummy_api_key,
            "limit": 200,
            "page": 1,
            "key1": "value1",
            "key2": "value2",
        }

        mock_requests_get.assert_called_with(self.dummy_url, params=expected_request)



import logging
import requests
from lastipy.lastfm.library.top_track import TopTrack
from requests import RequestException

URL = "http://ws.audioscrobbler.com/2.0/?method=track.getInfo"
MAX_RETRIES = 10


# TODO rename - track_info kinda vague
def fetch_playcount(track, user, api_key):
    retries = 0
    while retries <= MAX_RETRIES:
        try:
            json_response = _send_request(_build_payload(track, user, api_key))
            playcount = int(json_response["track"]["userplaycount"])
            logging.debug(
                user + " has played " + str(track) + " " + str(playcount) + " times"
            )
            return playcount
        except RequestException:
            if retries < MAX_RETRIES:
                logging.warning(
                    "Failed to fetch playcount for track "
                    + str(track)
                    + ". Retrying..."
                )
                retries += 1
            else:
                logging.warning(
                    "Failed to fetch playcount after "
                    + str(retries)
                    + " retries. Returning a playcount of 0."
                )
                return 0


def _send_request(json_payload):
    response = requests.get(URL, params=json_payload)
    if response.ok:
        return response.json()
    else:
        response.raise_for_status()


def _build_payload(track, user, api_key):
    payload = {
        "username": user,
        "api_key": api_key,
        "format": "json",
        "track": track.track_name,
        "artist": track.artist,
        "autocorrect": 1,
    }
    return payload



from lastipy.track import Track


class RecommendedTrack(Track):
    """Represents a recommended track, with a 'rating' value that can be used to determine it's 'strength'"""

    def __init__(self, track_name, artist, spotify_id=None, recommendation_rating=1):
        super().__init__(track_name, artist, spotify_id)
        self.recommendation_rating = recommendation_rating

    def __eq__(self, other):
        return (
            isinstance(other, RecommendedTrack)
            and self.track_name == other.track_name
            and self.artist == other.artist
            and self.spotify_id == other.spotify_id
            and self.recommendation_rating == other.recommendation_rating
        )

    def __hash__(self):
        return hash((self.track_name, self.artist, self.recommendation_rating))

    def __repr__(self):
        return str(self.__dict__)



import logging
from lastipy.spotify.parse_spotify_tracks import parse_tracks
from lastipy.util.chunk import chunk_list

MAX_ITEMS_PER_REQUEST = 100


# TODO test
def get_tracks_in_playlists(spotify):
    """Returns all tracks in the currently-logged-in user's playlists"""

    logging.info(
        "Fetching all tracks in " + spotify.current_user()["id"] + "'s playlists"
    )

    playlists = spotify.current_user_playlists()["items"]

    tracks = []
    for playlist in playlists:
        tracks = tracks + get_tracks_in_playlist(spotify, playlist_id=playlist["id"])

    logging.info(
        "Fetched "
        + str(len(tracks))
        + " tracks in "
        + str(len(playlists))
        + " playlists"
    )
    logging.debug("Fetched tracks: " + str(tracks))

    return tracks


# TODO test
def replace_tracks_in_playlist(spotify, playlist_name, tracks):
    """Replaces all tracks in the currently-logged-in user's given playlist
    with the given tracks. If the playlist does not exist, creates it first."""

    logging.info(
        "Replacing "
        + str(len(tracks))
        + " tracks in "
        + spotify.current_user()["id"]
        + "'s playlist "
        + playlist_name
    )
    logging.debug("Replacing with tracks: " + str(tracks))

    playlist_id = _get_playlist(spotify, playlist_name)

    if playlist_id is None:
        playlist_id = _create_playlist(spotify, playlist_name)

    spotify.user_playlist_replace_tracks(
        user=spotify.current_user()["id"],
        playlist_id=playlist_id,
        tracks=[track.spotify_id for track in tracks],
    )

    logging.info("Finished replacing tracks")


# TODO test
def get_tracks_in_playlist(spotify, playlist_name=None, playlist_id=None):
    """Returns all tracks in the given user's given playlist"""

    if playlist_id is None:
        if playlist_name is not None:
            logging.info("Fetching tracks in playlist with name " + playlist_name)
            playlist_id = _get_playlist(spotify, playlist_name)
        else:
            raise Exception("playlist_name or playlist_id required")
    else:
        logging.info("Fetching tracks in playlist with ID " + playlist_id)

    tracks_in_playlist = []
    keep_fetching = True
    while keep_fetching:
        json_response = spotify.playlist_tracks(
            playlist_id=playlist_id, offset=len(tracks_in_playlist)
        )
        if json_response["items"]:
            tracks_in_playlist = tracks_in_playlist + parse_tracks(
                json_response["items"]
            )
        else:
            keep_fetching = False

    logging.info("Fetched " + str(len(tracks_in_playlist)) + " tracks")
    logging.debug("Fetched tracks: " + str(tracks_in_playlist))

    return tracks_in_playlist


# TODO test
def add_tracks_to_playlist(spotify, playlist_name, tracks):
    playlist_id = _get_playlist(spotify, playlist_name)

    if playlist_id is None:
        playlist_id = _create_playlist(spotify, playlist_name)

    logging.info(
        "Adding "
        + str(len(tracks))
        + " tracks to "
        + spotify.current_user()["id"]
        + "'s playlist "
        + playlist_name
    )
    logging.debug("Adding tracks: " + str(tracks))

    tracks_chunked = chunk_list(tracks, MAX_ITEMS_PER_REQUEST)
    for chunk in tracks_chunked:
        # Remove any tracks that don't have a Spotify ID
        chunk = [track for track in chunk if track.spotify_id is not None]
        if len(chunk) > 0:
            spotify.user_playlist_add_tracks(
                user=spotify.current_user()["id"],
                playlist_id=playlist_id,
                tracks=[track.spotify_id for track in chunk],
            )
        else:
            logging.debug("Chunk is empty, not adding any tracks")

    logging.info("Finished adding tracks")


# TODO test
def remove_tracks_from_playlist(spotify, playlist_name, tracks):
    logging.info(
        "Removing "
        + str(len(tracks))
        + " tracks from "
        + spotify.current_user()["id"]
        + "'s playlist "
        + playlist_name
    )
    logging.debug("Removing tracks: " + str(tracks))

    playlist_id = _get_playlist(spotify, playlist_name)

    if playlist_id is None:
        playlist_id = _create_playlist(spotify, playlist_name)

    tracks_chunked = chunk_list(tracks, MAX_ITEMS_PER_REQUEST)
    for chunk in tracks_chunked:
        # Remove any tracks that don't have a Spotify ID
        chunk = [track for track in chunk if track.spotify_id is not None]
        if len(chunk) > 0:
            spotify.user_playlist_remove_all_occurrences_of_tracks(
                user=spotify.current_user()["id"],
                playlist_id=playlist_id,
                tracks=[track.spotify_id for track in chunk],
            )
        else:
            logging.debug("Chunk is empty, not removing any tracks")

    logging.debug("Finished removing tracks")


def _create_playlist(spotify, playlist_name):
    logging.info("Creating playlist " + playlist_name)
    return spotify.user_playlist_create(
        user=spotify.current_user()["id"], name=playlist_name
    )["id"]


def _get_playlist(spotify, playlist_name):
    logging.debug("Looking for playlist " + playlist_name)
    playlists = spotify.current_user_playlists()
    playlists = [
        playlist for playlist in playlists["items"] if playlist["name"] == playlist_name
    ]
    if playlists:
        return playlists[0]["id"]
    else:
        return None



from configparser import ConfigParser


class ApiKeysParser:
    def __init__(self, file):
        self._extract_keys(file)

    def _extract_keys(self, file):
        config_parser = ConfigParser()
        config_parser.read(file.name)
        self.lastfm_api_key = config_parser["LastFM"]["API"]
        spotify_section = "Spotify"
        self.spotify_client_id_key = config_parser[spotify_section]["CLIENT_ID"]
        self.spotify_client_secret_key = config_parser[spotify_section]["CLIENT_SECRET"]



import unittest
from unittest.mock import patch
from lastipy.lastfm.library import recent_tracks
from unittest.mock import Mock
from requests import HTTPError
from lastipy.track import Track


class RecentArtistsTest(unittest.TestCase):

    dummy_user = "dummyUser"
    dummy_api_key = "123456789"

    @patch("lastipy.lastfm.library.recent_tracks.fetch_paginated_response")
    def test_fetch(self, mock_paginated_endpoint_fetch):
        expected_tracks = [
            Track(track_name="Strawberry Fields Forever", artist="The Beatles"),
            Track(track_name="Badge", artist="Cream"),
            Track(track_name="Black Dog", artist="Led Zeppelin"),
        ]

        mock_paginated_endpoint_fetch.return_value = [
            {
                "recenttracks": {
                    "track": [
                        {
                            "name": "Strawberry Fields Forever",
                            "artist": {"name": "The Beatles"},
                        },
                        {"name": "Badge", "artist": {"name": "Cream"}},
                    ],
                    "@attr": {"totalPages": "2"},
                }
            },
            {
                "recenttracks": {
                    "track": [
                        {"name": "Black Dog", "artist": {"name": "Led Zeppelin"}}
                    ],
                    "@attr": {"totalPages": "2"},
                }
            },
        ]

        fetched_tracks = recent_tracks.fetch_recent_tracks(
            user=self.dummy_user, api_key=self.dummy_api_key
        )

        self.assertCountEqual(fetched_tracks, expected_tracks)



import unittest
from unittest.mock import patch

from lastipy.lastfm.library.scrobbled_artist import ScrobbledArtist
from lastipy.lastfm.library.top_track import TopTrack
from lastipy.recommendations.rating_calculator import calculate_ratings
from lastipy.recommendations.recommended_track import RecommendedTrack


class RatingCalculatorTest(unittest.TestCase):
    @patch("lastipy.recommendations.rating_calculator.fetch_recent_artists")
    def test_rating_with_one_top_track_one_recommendation(self, mock_recent_artists):
        mock_recent_artists.return_value = []
        top_track = TopTrack(
            artist="The Beatles", track_name="Penny Lane", playcount=10
        )
        recommended_track_1 = RecommendedTrack(
            artist="The Beatles",
            track_name="Strawberry Fields Forever",
            recommendation_rating=1,
        )
        top_tracks_to_recommendations = {top_track: [recommended_track_1]}
        recommendations = calculate_ratings("test", "", top_tracks_to_recommendations)
        self.assertEqual(10, recommendations[0].recommendation_rating)

    @patch("lastipy.recommendations.rating_calculator.fetch_recent_artists")
    def test_rating_with_one_top_track_multiple_recommendations(
        self, mock_recent_artists
    ):
        mock_recent_artists.return_value = []
        top_track = TopTrack(
            artist="The Beatles", track_name="Penny Lane", playcount=10
        )
        recommended_track_1 = RecommendedTrack(
            artist="The Beatles",
            track_name="Strawberry Fields Forever",
            recommendation_rating=1,
        )
        recommended_track_2 = RecommendedTrack(
            artist="The Beatles",
            track_name="Baby You're a Rich Man",
            recommendation_rating=1,
        )
        recommended_track_3 = RecommendedTrack(
            artist="The Beatles", track_name="Blue Jay Way", recommendation_rating=1
        )
        top_tracks_to_recommendations = {
            top_track: [recommended_track_1, recommended_track_2, recommended_track_3]
        }
        recommendations = calculate_ratings("test", "", top_tracks_to_recommendations)
        self.assertEqual(10, recommendations[0].recommendation_rating)
        self.assertEqual(10, recommendations[1].recommendation_rating)
        self.assertEqual(10, recommendations[2].recommendation_rating)

    @patch("lastipy.recommendations.rating_calculator.fetch_recent_artists")
    def test_rating_with_multiple_top_tracks_multiple_recommendations(
        self, mock_recent_artists
    ):
        mock_recent_artists.return_value = []
        top_track_1 = TopTrack(
            artist="The Beatles", track_name="Penny Lane", playcount=10
        )
        recommended_track_1 = RecommendedTrack(
            artist="The Beatles",
            track_name="Strawberry Fields Forever",
            recommendation_rating=1,
        )
        recommended_track_2 = RecommendedTrack(
            artist="The Beatles",
            track_name="Baby You're a Rich Man",
            recommendation_rating=1,
        )
        recommended_track_3 = RecommendedTrack(
            artist="The Beatles", track_name="Blue Jay Way", recommendation_rating=1
        )
        top_track_2 = TopTrack(
            artist="The Rolling Stones", track_name="Satisfaction", playcount=20
        )
        recommended_track_4 = RecommendedTrack(
            artist="The Rolling Stones",
            track_name="Ruby Tuesday",
            recommendation_rating=1,
        )
        recommended_track_5 = RecommendedTrack(
            artist="The Beatles",
            track_name="Get Off Of My Cloud",
            recommendation_rating=1,
        )
        top_tracks_to_recommendations = {
            top_track_1: [
                recommended_track_1,
                recommended_track_2,
                recommended_track_3,
            ],
            top_track_2: [recommended_track_4, recommended_track_5],
        }
        recommendations = calculate_ratings("test", "", top_tracks_to_recommendations)
        self.assertEqual(10, recommendations[0].recommendation_rating)
        self.assertEqual(10, recommendations[1].recommendation_rating)
        self.assertEqual(10, recommendations[2].recommendation_rating)
        self.assertEqual(20, recommendations[3].recommendation_rating)
        self.assertEqual(20, recommendations[4].recommendation_rating)

    @patch("lastipy.recommendations.rating_calculator.fetch_recent_artists")
    def test_rating_is_reduced_based_on_artist_playcount(self, mock_recent_artists):
        mock_recent_artists.return_value = [
            ScrobbledArtist(artist_name="The Beatles", playcount=9)
        ]
        top_track = TopTrack(artist="The Beatles", track_name="Penny Lane", playcount=5)
        recommended_track_1 = RecommendedTrack(
            artist="The Beatles",
            track_name="Strawberry Fields Forever",
            recommendation_rating=1,
        )
        recommended_track_2 = RecommendedTrack(
            artist="The Beatles", track_name="Blue Jay Way", recommendation_rating=1
        )
        top_tracks_to_recommendations = {
            top_track: [recommended_track_1, recommended_track_2]
        }
        recommendations = calculate_ratings("test", "", top_tracks_to_recommendations)
        self.assertEqual(5 / (9), recommendations[0].recommendation_rating)
        self.assertEqual(5 / (9), recommendations[1].recommendation_rating)

    @patch("lastipy.recommendations.rating_calculator.fetch_recent_artists")
    def test_multiple_recommendations_and_multiple_recent_artists(
        self, mock_recent_artists
    ):
        mock_recent_artists.return_value = [
            ScrobbledArtist(artist_name="The Beatles", playcount=9),
            ScrobbledArtist(artist_name="The Rolling Stones", playcount=19),
        ]
        top_track_1 = TopTrack(
            artist="The Beatles", track_name="Penny Lane", playcount=5
        )
        recommended_track_1 = RecommendedTrack(
            artist="The Beatles",
            track_name="Strawberry Fields Forever",
            recommendation_rating=1,
        )
        top_track_2 = TopTrack(
            artist="The Rolling Stones", track_name="Satisfaction", playcount=10
        )
        recommended_track_2 = RecommendedTrack(
            artist="The Kinks", track_name="You Really Got Me", recommendation_rating=1
        )
        recommended_track_3 = RecommendedTrack(
            artist="The Rolling Stones",
            track_name="Ruby Tuesday",
            recommendation_rating=1,
        )
        top_tracks_to_recommendations = {
            top_track_1: [recommended_track_1],
            top_track_2: [recommended_track_2, recommended_track_3],
        }
        recommendations = calculate_ratings("test", "", top_tracks_to_recommendations)
        self.assertEqual(5 / 9, recommendations[0].recommendation_rating)
        self.assertEqual(10, recommendations[1].recommendation_rating)
        self.assertEqual(10 / 19, recommendations[2].recommendation_rating)

    @patch("lastipy.recommendations.rating_calculator.fetch_recent_artists")
    def test_reducing_based_on_recent_artists_ignores_case(self, mock_recent_artists):
        mock_recent_artists.return_value = [
            ScrobbledArtist(artist_name="LANY", playcount=3)
        ]
        top_track = TopTrack(artist="LANY", track_name="Some Whiny Crap", playcount=2)
        recommended_track = RecommendedTrack(
            artist="Lany", track_name="Some Other Whiny Crap", recommendation_rating=1
        )
        top_tracks_to_recommendations = {top_track: [recommended_track]}
        recommendations = calculate_ratings("test", "", top_tracks_to_recommendations)
        self.assertEqual(2 / 3, recommendations[0].recommendation_rating)



import unittest
from lastipy.lastfm.library import period
from lastipy.lastfm.library.top_track import TopTrack
from unittest.mock import patch, Mock
from requests import HTTPError
from lastipy.lastfm.library.top_tracks import fetch_top_tracks


class TopTracksFetcherTest(unittest.TestCase):

    dummy_user = "dummyUser"
    dummy_api_key = "123456789"

    @patch("lastipy.lastfm.library.top_tracks.fetch_paginated_response")
    def test_one_page_of_results(self, mock_paginated_endpoint_fetch):
        expected_track_1 = TopTrack(
            track_name="Stayin Alive", artist="Bee Gees", playcount=2
        )
        expected_track_2 = TopTrack(track_name="Badge", artist="Cream", playcount=7)

        json_response_page_1 = {
            "toptracks": {
                "track": [
                    {
                        "name": "Stayin Alive",
                        "artist": {"name": "Bee Gees"},
                        "playcount": "2",
                    }
                ]
            }
        }

        json_response_page_2 = {
            "toptracks": {
                "track": [
                    {"name": "Badge", "artist": {"name": "Cream"}, "playcount": "7"}
                ]
            }
        }

        mock_paginated_endpoint_fetch.return_value = [
            json_response_page_1,
            json_response_page_2,
        ]

        fetched_tracks = fetch_top_tracks(
            user=self.dummy_user, api_key=self.dummy_api_key, a_period=period.ONE_MONTH
        )
        self.assertCountEqual(fetched_tracks, [expected_track_1, expected_track_2])

    @patch("lastipy.lastfm.library.top_tracks.fetch_paginated_response")
    def test_songs_with_one_playcount_ignored(self, mock_paginated_endpoint_fetch):
        json_response_page = {
            "toptracks": {
                "track": [
                    {
                        "name": "Stayin Alive",
                        "artist": {"name": "Bee Gees"},
                        "playcount": "1",
                    }
                ]
            }
        }

        mock_paginated_endpoint_fetch.return_value = [json_response_page]

        fetched_tracks = fetch_top_tracks(
            user=self.dummy_user, api_key=self.dummy_api_key, a_period=period.ONE_MONTH
        )
        self.assertCountEqual(fetched_tracks, [])



import spotipy
from lastipy.spotify.parse_spotify_tracks import parse_tracks
from lastipy.spotify.library import get_saved_tracks
from lastipy.spotify.playlist import get_tracks_in_playlists
from datetime import datetime
import logging
from lastipy.track import Track
from lastipy.spotify import album
from lastipy.util.filter import filter_out_duplicates, filter_out_tracks_in_second_list


# TODO test
def fetch_new_tracks(
    spotify,
    album_types=[album.SINGLE_ALBUM_TYPE, album.ALBUM_ALBUM_TYPE],
    ignore_remixes=False,
    ignore_songs_in_library=True,
    as_of_date=datetime.today().date(),
):
    """Fetches new tracks (as of the given date) released by the current Spotify user's followed artists"""

    logging.info(
        "Fetching new tracks for "
        + spotify.current_user()["id"]
        + " as of "
        + str(as_of_date)
    )

    new_albums = fetch_new_albums(spotify, album_types, as_of_date)

    all_tracks = []
    for album in new_albums:
        all_tracks += _fetch_album_tracks(spotify, album)

    new_tracks = parse_tracks(all_tracks)
    new_tracks = filter_out_duplicates(new_tracks)

    if ignore_remixes:
        logging.info("Filtering out those pesky remixes...")
        new_tracks = [
            track for track in new_tracks if "remix" not in track.track_name.lower()
        ]

    if ignore_songs_in_library:
        logging.info(
            "Filtering out tracks that are already in the user's saved tracks and playlists..."
        )
        saved_tracks = get_saved_tracks(spotify)
        new_tracks = filter_out_tracks_in_second_list(new_tracks, saved_tracks)

        playlist_tracks = get_tracks_in_playlists(spotify)
        new_tracks = filter_out_tracks_in_second_list(new_tracks, playlist_tracks)

    logging.info("Fetched " + str(len(new_tracks)) + " new tracks " + str(new_tracks))
    return new_tracks


# TODO test
def fetch_new_albums(
    spotify,
    album_types=[album.SINGLE_ALBUM_TYPE, album.ALBUM_ALBUM_TYPE],
    as_of_date=datetime.today().date(),
):
    """Fetches new albums (as of the given date) released by the given Spotify user's followed artists"""

    followed_artist_ids = _fetch_followed_artists(spotify)

    logging.info(
        "Fetching new albums for "
        + spotify.current_user()["id"]
        + " as of "
        + str(as_of_date)
    )

    all_albums = []
    for artist_id in followed_artist_ids:
        artist_albums = _fetch_artist_albums(spotify, album_types, artist_id)
        all_albums += artist_albums

    # TODO remove this once https://github.com/evanjamesjackson/lastipy/issues/22 is fixed...
    logging.info(
        "Before filtering by date, fetched " + str(len(all_albums)) + " albums"
    )
    logging.info(str(all_albums))

    new_albums = _filter_new_albums(all_albums, as_of_date)

    logging.info("Fetched " + str(len(new_albums)) + " new albums " + str(new_albums))
    return new_albums


def _fetch_followed_artists(spotify):
    followed_artists = []

    curr_response = spotify.current_user_followed_artists(limit=50)

    while len(curr_response["artists"]["items"]) > 0:
        curr_response = spotify.current_user_followed_artists(
            limit=50,
            after=curr_response["artists"]["items"][len(curr_response) - 1]["id"],
        )
        followed_artists += curr_response["artists"]["items"]

    # The above Spotipy function doesn't really seem to function properly and results in duplicates,
    # so we remove them here by converting the list to just the IDs (not doing so results in
    # an unhashable error), then converting to a set and back to a list
    followed_artists = [artist["id"] for artist in followed_artists]
    followed_artist_ids = list(set(followed_artists))

    logging.debug("Fetched " + str(len(followed_artist_ids)) + " followed artists " + str(followed_artist_ids))

    return followed_artist_ids


def _filter_new_albums(all_albums, as_of_date):
    new_albums = []
    for album in all_albums:
        if album.release_date_precision == "day":
            if datetime.strptime(album.release_date, "%Y-%m-%d").date() == as_of_date:
                new_albums.append(album)
        else:
            logging.warn(
                "Album release date precision is not 'day' so ignoring (album: "
                + str(album)
                + ")"
            )
    return new_albums


def _fetch_artist_albums(spotify, album_types, artist_id):
    logging.debug("Fetching albums with types " + str(album_types) + " for artist ID " + str(artist_id))
    albums = []
    for album_type in album_types:
        curr_response = _artist_albums(
            spotify=spotify, artist_id=artist_id, album_types=album_type, limit=50
        )
        albums += _convert_albums(curr_response)
        while len(curr_response["items"]) > 0:
            curr_response = _artist_albums(
                spotify=spotify, artist_id=artist_id, album_types=album_type, limit=50, offset=len(albums)
            )
            albums += _convert_albums(curr_response)
    logging.debug("Fetched " + str(len(albums)) + " albums for artist " + str(artist_id) + " " + str(albums))
    return albums


def _convert_albums(json_album_response):
    return [
        album.SpotifyAlbum(
            name=item["name"],
            artist=item["artists"][0]["name"],
            album_type=item["album_type"],
            spotify_id=item["id"],
            release_date_precision=item["release_date_precision"],
            release_date=item["release_date"],
        )
        for item in json_album_response["items"]
    ]


def _fetch_album_tracks(spotify, album):
    curr_response = spotify.album_tracks(album.spotify_id, limit=50)
    album_tracks = curr_response["items"]
    while len(curr_response["items"]) > 0:
        curr_response = spotify.album_tracks(
            album.spotify_id, limit=50, offset=len(album_tracks)
        )
        album_tracks += curr_response["items"]
    return album_tracks

# Modified version of this method from Spotipy until https://github.com/spotipy-dev/spotipy/pull/1108 is included in a release
def _artist_albums(
    spotify, artist_id, album_types=None, country=None, limit=20, offset=0
):
    trid = spotify._get_id("artist", artist_id)
    return spotify._get(
        "artists/" + trid + "/albums",
        include_groups=album_types,
        country=country,
        limit=limit,
        offset=offset,
    )


import logging
from lastipy.track import Track


# TODO test
def filter_out_duplicates(tracks):
    """Filters out duplicate tracks from the given list (i.e. tracks that have the same track name and artist; see Track.are_equivalent)"""

    logging.info("Filtering out duplicates")
    logging.debug(str(len(tracks)) + " tracks before removing duplicates")
    tracks_without_duplicates = []
    # We remove duplicates with a list comprehension rather than the traditional hack of using a set, since that
    # requires the object to be hashable; plus we only want to compare the track name/artist of each track, not
    # any of the other fields (eg: Spotify ID) which might in fact differ
    [
        tracks_without_duplicates.append(track_x)
        for track_x in tracks
        if not any(
            Track.are_equivalent(track_x, track_y)
            for track_y in tracks_without_duplicates
        )
    ]
    logging.debug(str(len(tracks)) + " tracks after removing duplicates")
    return tracks_without_duplicates


# TODO test
def filter_out_tracks_in_second_list(track_list_1, track_list_2):
    """Filters out tracks from track_list_1 that are also in track_list_2"""
    filtered_tracks = [
        track_x
        for track_x in track_list_1
        if not any(Track.are_equivalent(track_x, track_y) for track_y in track_list_2)
    ]
    return filtered_tracks



from lastipy.spotify.parse_spotify_tracks import parse_tracks
from lastipy.spotify.playlist import get_tracks_in_playlist
import spotipy
import logging
from lastipy.util.chunk import chunk_list

MAX_ITEMS_PER_REQUEST = 50


def get_saved_tracks(spotify):
    """Returns the current logged-in users's saved tracks (Liked Songs)"""

    logging.info("Fetching " + spotify.current_user()["id"] + "'s saved tracks")

    saved_tracks = []
    keep_fetching = True
    while keep_fetching:
        json_response = spotify.current_user_saved_tracks(offset=len(saved_tracks))
        if json_response["items"]:
            saved_tracks = saved_tracks + parse_tracks(json_response["items"])
        else:
            keep_fetching = False

    logging.info("Fetched " + str(len(saved_tracks)) + " saved tracks")
    logging.debug("Fetched tracks: " + str(saved_tracks))

    return saved_tracks


def add_tracks_to_library(spotify, tracks):
    """Adds the given tracks to the current logged-in user's saved tracks (Liked Songs)"""

    logging.info(
        "Adding "
        + str(len(tracks))
        + " to "
        + spotify.current_user()["id"]
        + "'s library"
    )
    logging.debug("Adding tracks: " + str(tracks))
    track_chunks = chunk_list(tracks, MAX_ITEMS_PER_REQUEST)
    for chunk in track_chunks:
        spotify.current_user_saved_tracks_add([track.spotify_id for track in chunk])
    logging.info("Finished adding tracks")


def add_albums_to_library(spotify, albums):
    """Adds the given albums to the current logged-in user's library"""

    logging.info(
        "Adding "
        + str(len(albums))
        + " to "
        + spotify.current_user()["id"]
        + "'s library"
    )
    logging.debug("Adding albums: " + str(albums))
    album_chunks = chunk_list(albums, MAX_ITEMS_PER_REQUEST)
    for chunk in album_chunks:
        spotify.current_user_saved_albums_add([album.spotify_id for album in chunk])
    logging.info("Finished adding albums")


# TODO test
def remove_tracks_from_library(spotify, tracks):
    """Removes the given tracks from the current logged-in user's saved tracks (Liked Songs)"""

    logging.info(
        "Removing "
        + str(len(tracks))
        + " from "
        + spotify.current_user()["id"]
        + "'s library"
    )
    logging.debug("Removing tracks: " + str(tracks))
    track_chunks = chunk_list(tracks, MAX_ITEMS_PER_REQUEST)
    for chunk in track_chunks:
        spotify.current_user_saved_tracks_delete([track.spotify_id for track in chunk])
    logging.info("Finished removing tracks")



import unittest
from unittest.mock import patch

from lastipy.recommendations.recommendations import generate_recommendations
from lastipy.lastfm.library.top_track import TopTrack
from lastipy.recommendations.recommended_track import RecommendedTrack


class RecommendationsGeneratorTest(unittest.TestCase):
    @patch("lastipy.recommendations.recommendations.fetch_recent_tracks")
    @patch("lastipy.recommendations.recommendations.fetch_top_tracks")
    @patch("lastipy.recommendations.recommendations.lastfm_recommendations")
    @patch("lastipy.recommendations.recommendations.calculate_ratings")
    @patch("lastipy.recommendations.recommendations.library")
    @patch("lastipy.recommendations.recommendations.playlist")
    def test_recent_tracks_are_filtered(
        self,
        mock_spotify_playlist,
        mock_spotify_library,
        mock_calculate_ratings,
        mock_fetch_recommendations,
        mock_top_tracks,
        mock_recent_tracks,
    ):
        recent_track_1 = TopTrack(track_name="SWALBR", artist="Cream", playcount=1)
        recent_track_2 = TopTrack(track_name="Badge", artist="Cream", playcount=1)
        recent_track_3 = TopTrack(
            track_name="Layla", artist="Derek & the Dominos", playcount=8
        )
        mock_recent_tracks.return_value = [
            recent_track_1,
            recent_track_2,
            recent_track_3,
        ]

        mock_top_tracks.return_value = [recent_track_3]

        new_recommendation = RecommendedTrack(
            track_name="Key to the Highway",
            artist="Derek & the Dominos",
            recommendation_rating=8,
        )
        already_scrobbled_recommendation_1 = RecommendedTrack(
            track_name="Badge", artist="Cream", recommendation_rating=8
        )
        already_scrobbled_recommendation_2 = RecommendedTrack(
            track_name="SWALBR", artist="Cream", recommendation_rating=8
        )
        recommendations = [
            new_recommendation,
            already_scrobbled_recommendation_1,
            already_scrobbled_recommendation_2,
        ]
        mock_fetch_recommendations.fetch_recommendations.return_value = recommendations

        mock_calculate_ratings.return_value = recommendations
        mock_spotify_library.get_saved_tracks.return_value = []
        mock_spotify_playlist.get_tracks_in_playlists.return_value = []

        self.assertCountEqual(
            generate_recommendations(
                spotify=None, lastfm_user="test", lastfm_api_key=""
            ),
            [new_recommendation],
        )

    @patch("lastipy.recommendations.recommendations.fetch_recent_tracks")
    @patch("lastipy.recommendations.recommendations.fetch_top_tracks")
    @patch("lastipy.recommendations.recommendations.lastfm_recommendations")
    @patch("lastipy.recommendations.recommendations.calculate_ratings")
    @patch("lastipy.recommendations.recommendations.library")
    @patch("lastipy.recommendations.recommendations.playlist")
    def test_blacklisted_artists_are_filtered(
        self,
        mock_spotify_playlist,
        mock_spotify_library,
        mock_calculate_ratings,
        mock_fetch_recommendations,
        mock_top_tracks,
        mock_recent_tracks,
    ):
        mock_recent_tracks.return_value = []
        mock_top_tracks.return_value = [
            TopTrack(track_name="Here Comes the Sun", artist="The Beatles", playcount=5)
        ]

        recommendation_1 = RecommendedTrack(
            track_name="Badge", artist="Cream", recommendation_rating=1
        )
        recommendation_2 = RecommendedTrack(
            track_name="Penny Lane", artist="The Beatles", recommendation_rating=1
        )
        recommendations = [recommendation_1, recommendation_2]
        mock_fetch_recommendations.fetch_recommendations.return_value = recommendations

        mock_calculate_ratings.return_value = recommendations
        mock_spotify_library.get_saved_tracks.return_value = []
        mock_spotify_playlist.get_tracks_in_playlists.return_value = []

        recommendations = generate_recommendations(
            lastfm_user="meeee",
            lastfm_api_key="",
            spotify=None,
            blacklisted_artists=["The Beatles"],
        )

        self.assertCountEqual(recommendations, [recommendation_1])

    @patch("lastipy.recommendations.recommendations.fetch_recent_tracks")
    @patch("lastipy.recommendations.recommendations.fetch_top_tracks")
    @patch("lastipy.recommendations.recommendations.lastfm_recommendations")
    @patch("lastipy.recommendations.recommendations.calculate_ratings")
    @patch("lastipy.recommendations.recommendations.library")
    @patch("lastipy.recommendations.recommendations.playlist")
    def test_blacklisted_artists_filtering_should_ignore_case(
        self,
        mock_spotify_playlist,
        mock_spotify_library,
        mock_calculate_ratings,
        mock_fetch_recommendations,
        mock_top_tracks,
        mock_recent_tracks,
    ):
        mock_recent_tracks.return_value = []
        mock_top_tracks.return_value = [
            TopTrack(
                track_name="everything i wanted", artist="Billie Eilish", playcount=5
            )
        ]

        recommendation = RecommendedTrack(
            track_name="Bad Music", artist="Zayn", recommendation_rating=1
        )
        recommendations = [recommendation]
        mock_fetch_recommendations.fetch_recommendations.return_value = recommendations

        mock_calculate_ratings.return_value = recommendations
        mock_spotify_library.get_saved_tracks.return_value = []
        mock_spotify_playlist.get_tracks_in_playlists.return_value = []

        recommendations = generate_recommendations(
            lastfm_user="",
            lastfm_api_key="",
            spotify=None,
            blacklisted_artists=["ZAYN"],
        )

        self.assertEqual(0, len(recommendations))



def parse_track_name(json_track):
    return json_track["name"]


def parse_artist(json_track):
    artist_json = json_track["artist"]
    if "name" in artist_json:
        artist = artist_json["name"]
    elif "#text" in artist_json:
        artist = artist_json["#text"]
    return artist



from setuptools import setup, find_packages

setup(
    name="lastipy",
    # fmt: off
    version='1.1.43',
    # fmt: on
    description="Lastipy is a Python library combining the APIs of Spotify and Last.fm, with scripts for creating customized recommendation playlists, automatically saving new releases, etc.",
    url="http://github.com/evanjamesjackson/lastipy",
    author="Evan Jackson",
    author_email="evan@jacksonnet.ca",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "recommendations_playlist = scripts.recommendations_playlist:build_recommendations_playlist",
            "save_new_releases = scripts.save_new_releases:save_new_releases",
            "organize_favorites = scripts.organize_favorites:organize_favorites",
        ]
    },
    install_requires=[
        "numpy",
        "requests",
        "spotipy",
        "iso8601",
        "python-dateutil",
        "pytest",
        "wheel",
        "black"
    ]
)



import logging
import requests
from lastipy.lastfm.library.scrobbled_artist import ScrobbledArtist
from lastipy.lastfm.library.paginated_endpoint import fetch_paginated_response

URL = "http://ws.audioscrobbler.com/2.0/?method=library.getartists"


def fetch_recent_artists(user, api_key):
    """Fetches recent artists for the given user"""

    logging.info("Fetching recent artists for user " + user)

    paginated_json_responses = fetch_paginated_response(URL, user, api_key, "artists")

    artists = []
    for json_response in paginated_json_responses:
        # Weirdly, the artist attribute is sometimes a list, and sometimes an object (if there is only one artist in the response)
        if isinstance(json_response["artists"]["artist"], list):
            for json_artist in json_response["artists"]["artist"]:
                artists.append(_build_artist(json_artist))
        else:
            json_artist = json_response["artists"]["artist"]
            artists.append(_build_artist(json_artist))

    logging.info("Fetched " + str(len(artists)) + " artists")
    logging.debug("Fetched artists: " + str(artists))
    return artists


def _build_artist(artist):
    return ScrobbledArtist(
        artist_name=artist["name"], playcount=int(artist["playcount"])
    )



from lastipy import definitions
import logging
import logging.handlers
import os


def setup_logging(filename):
    logs_directory = os.path.join(definitions.ROOT_DIR, "logs")
    if not os.path.exists(logs_directory):
        os.makedirs(logs_directory)

    logger = logging.getLogger()
    formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
    logger.setLevel(logging.DEBUG)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)
    console_handler.setFormatter(formatter)

    file_handler = logging.handlers.RotatingFileHandler(
        filename=os.path.join(logs_directory, filename),
        maxBytes=2 * 1024 * 1024,  # 2MB
        backupCount=25,
        encoding="utf-8",
    )
    file_handler.setFormatter(formatter)
    file_handler.setLevel(logging.INFO)

    logger.addHandler(console_handler)
    logger.addHandler(file_handler)



from lastipy.track import Track
from lastipy.spotify.playlist_track import PlaylistTrack

# TODO test
def parse_tracks(json_tracks):
    return [_parse_track(json_track) for json_track in json_tracks]


def _parse_track(json_track):
    if "added_at" in json_track:
        return _parse_playlist_track(json_track)
    else:
        name, artist, track_id = _parse_common_track_properties(json_track)
        return Track(track_name=name, artist=artist, spotify_id=track_id)


def _parse_playlist_track(json_track):
    added_at = json_track["added_at"]
    if "track" in json_track:
        json_track = json_track["track"]
    name, artist, track_id = _parse_common_track_properties(json_track)
    return PlaylistTrack(
        track_name=name, artist=artist, spotify_id=track_id, added_at=added_at
    )


def _parse_common_track_properties(json_track):
    track_id = json_track["id"]
    name = json_track["name"]
    # Just getting the first artist, even if there's multiple
    artist = json_track["artists"][0]["name"]
    return name, artist, track_id



import logging

from lastipy.lastfm.library.top_tracks import fetch_top_tracks
from lastipy.lastfm import lastfm_recommendations
from lastipy.spotify import spotify_recommendations
from lastipy.track import Track
from lastipy.lastfm.library.recent_tracks import fetch_recent_tracks
from lastipy.recommendations.rating_calculator import calculate_ratings
from lastipy.spotify import library, search, playlist
from lastipy.lastfm.library import period
from numpy.random import choice
from lastipy.util.filter import filter_out_duplicates, filter_out_tracks_in_second_list


def generate_recommendations(
    lastfm_user,
    lastfm_api_key,
    spotify,
    recommendation_services="Last.fm",
    recommendation_period=period.OVERALL,
    max_recommendations_per_top_track=100,
    blacklisted_artists=[],
    prefer_unheard_artists=True,
):
    """Fetches recommendations for the given user based on their top tracks in Last.fm"""

    logging.info(
        "Fetching recommendations based on " + lastfm_user + "'s top tracks in Last.fm"
    )

    top_tracks = fetch_top_tracks(
        user=lastfm_user, api_key=lastfm_api_key, a_period=recommendation_period
    )

    top_tracks_to_recommendations = {}
    recommendations = []
    for top_track in top_tracks:
        try:
            recommendations_for_current_track = _fetch_recommendations(
                recommendation_services,
                lastfm_api_key,
                spotify,
                top_track,
                max_recommendations_per_top_track,
            )
            if recommendations_for_current_track:
                recommendations = recommendations + recommendations_for_current_track
                top_tracks_to_recommendations[
                    top_track
                ] = recommendations_for_current_track
        except Exception as e:
            logging.error(
                f"An error occurred fetching recommendations for track "
                + str(top_track)
                + ": "
                + str(e)
            )

    recommendations = calculate_ratings(
        user=lastfm_user,
        api_key=lastfm_api_key,
        prefer_unheard_artists=prefer_unheard_artists,
        top_tracks_to_recommendations=top_tracks_to_recommendations,
    )

    logging.debug(
        f"Before filtering, fetched "
        + str(len(recommendations))
        + " recommendations: "
        + str(recommendations)
    )
    recommendations = _filter_out_recent_tracks(
        lastfm_user, lastfm_api_key, recommendations
    )
    recommendations = _filter_out_blacklisted_artists(
        blacklisted_artists, recommendations
    )
    recommendations = _filter_out_saved_tracks(recommendations, spotify)
    recommendations = _filter_out_playlist_tracks(recommendations, spotify)
    recommendations = filter_out_duplicates(recommendations)
    logging.info(
        "After filtering, fetched " + str(len(recommendations)) + " recommendations"
    )
    logging.debug("Recommendations: " + str(recommendations))

    return recommendations


def _fetch_recommendations(
    recommendation_services, lastfm_api_key, spotify, track, limit
):
    recommendations = []
    if "Last.fm" in recommendation_services:
        recommendations += lastfm_recommendations.fetch_recommendations(
            api_key=lastfm_api_key, limit=limit, track=track
        )
    if "Spotify" in recommendation_services:
        recommendations += spotify_recommendations.fetch_recommendations(
            spotify=spotify, track=track, limit=limit
        )
    return recommendations


def _filter_out_recent_tracks(user, api_key, recommendations):
    logging.info("Filtering out recent tracks from recommendations")
    recent_tracks = fetch_recent_tracks(user, api_key)
    return filter_out_tracks_in_second_list(recommendations, recent_tracks)


def _filter_out_blacklisted_artists(blacklisted_artists, recommendations):
    logging.info(
        "Filtering out blacklisted artists ("
        + str(blacklisted_artists)
        + ") from recommendations"
    )
    recommendations = [
        recommendation
        for recommendation in recommendations
        if not any(
            recommendation.artist.lower() == blacklisted_artist.lower()
            for blacklisted_artist in blacklisted_artists
        )
    ]
    return recommendations


# TODO test
def _filter_out_saved_tracks(recommendations, spotify):
    logging.info("Filtering out saved tracks from recommendations")
    saved_tracks = library.get_saved_tracks(spotify)
    return filter_out_tracks_in_second_list(recommendations, saved_tracks)


# TODO test
def _filter_out_playlist_tracks(recommendations, spotify):
    logging.info("Filtering out tracks in the user's playlists from recommendations")
    playlist_tracks = playlist.get_tracks_in_playlists(spotify)
    return filter_out_tracks_in_second_list(recommendations, playlist_tracks)



import unittest
from lastipy.lastfm.parse_lastfm_tracks import parse_artist, parse_track_name


class ParseTracksTest(unittest.TestCase):
    def test_parse_track_name_valid_json(self):
        track_name = "Stayin' Alive"
        artist = "Bee Gees"
        to_parse = {
            "name": track_name,
            "artist": {"name": artist},
        }

        parsed_track_name = parse_track_name(to_parse)

        self.assertEqual(parsed_track_name, track_name)

    def test_parse_artist_with_weird_artist_data(self):
        track_name = "Stayin' Alive"
        artist = "Bee Gees"
        to_convert = {
            "name": track_name,
            "artist": {"mbid": "45c25199-fa62-4d4c-b0a2-11eeed6923c3", "#text": artist},
        }

        parsed_artist = parse_artist(to_convert)

        self.assertEqual(parsed_artist, artist)



import logging
import spotipy
from lastipy.spotify import token
from lastipy.spotify.parse_spotify_tracks import parse_tracks
import json


# TODO test
def search_for_tracks(spotify, query):
    """Returns a list of tracks for the given query"""

    logging.info("Searching in Spotify with query " + query)

    json_response = spotify.search(q=query)
    logging.debug(
        "All search results for query " + query + ": " + json.dumps(json_response)
    )

    converted_tracks = []
    if json_response["tracks"]:
        converted_tracks = parse_tracks(json_response["tracks"]["items"])
    logging.info(
        "Found "
        + str(len(converted_tracks))
        + " tracks for search query "
        + query
        + " in Spotify"
    )
    logging.debug("Found tracks: " + str(converted_tracks))
    return converted_tracks



import unittest
from lastipy.lastfm.library.top_track import TopTrack
from lastipy.recommendations.recommended_track import RecommendedTrack
from unittest.mock import patch
from lastipy.lastfm.lastfm_recommendations import fetch_recommendations


class LastFmRecommendationsTest(unittest.TestCase):
    @patch("requests.get")
    def test_track_has_multiple_similar_tracks(self, mock_requests_get):
        playcount = 5
        track_to_check = TopTrack(
            track_name="Night Fever", artist="Bee Gees", playcount=playcount
        )
        expected_similar_tracks = [
            RecommendedTrack(
                track_name="Stayin' Alive", artist="Bee Gees", recommendation_rating=1
            ),
            RecommendedTrack(
                track_name="You Should Be Dancing",
                artist="Bee Gees",
                recommendation_rating=1,
            ),
        ]

        json_tracks = []
        for track in expected_similar_tracks:
            json_track = {
                "name": track.track_name,
                "artist": {"name": track.artist},
                "match": track.recommendation_rating,
            }
            json_tracks.append(json_track)

        json_response = {"similartracks": {"track": json_tracks}}

        mock_requests_get.return_value.json.return_value = json_response

        fetched_tracks = fetch_recommendations("", track_to_check, 2)
        self.assertCountEqual(expected_similar_tracks, fetched_tracks)



#!/usr/bin/env python3.7

from lastipy.spotify import library
from spotipy import Spotify
from lastipy.spotify import token
from configparser import ConfigParser
import argparse
import os
from lastipy import definitions
from lastipy.lastfm.library.top_tracks import fetch_top_tracks
from lastipy.lastfm.library.recent_artists import fetch_recent_artists
from lastipy.lastfm.library import period
from lastipy.track import Track
from numpy.random import choice
from spotipy import Spotify
from lastipy.spotify import token
from lastipy.util.setup_logging import setup_logging
import logging
from lastipy.util.parse_api_keys import ApiKeysParser
from lastipy.spotify import library, search, playlist
from lastipy.lastfm.library import track_info
from datetime import datetime, timezone
from dateutil.relativedelta import relativedelta
import iso8601


def organize_favorites():
    """Organizes a user's Spotify library. Saved tracks with more than args.saved_song_playount_limit plays are moved to a "New Favorites" playlist.
    Tracks in the "New Favorites" playlist are moved to an "Old Favorites" playlist after they've been in "New Favorites" for
    args.new_favorites_time_limit days, unless they have less than or equal to args.new_favorites_playcount_limit plays, in which case they're instead moved
    to a "Neglected" playlist."""

    setup_logging("organize_favorites.log")
    args = _extract_args()
    spotify = Spotify(
        auth=token.get_token(
            args.spotify_user,
            args.spotify_client_id_key,
            args.spotify_client_secret_key,
        )
    )

    logging.info("Organizing " + args.spotify_user + "'s favorites")
    move_saved_tracks(spotify, args)
    move_new_favorites(spotify, args)
    logging.info("Done!")


def move_new_favorites(spotify, args):
    new_favorites_tracks = playlist.get_tracks_in_playlist(
        spotify, playlist_name=args.new_favorites_playlist
    )
    old_favorites_tracks = []
    neglected_tracks = []
    for track in new_favorites_tracks:
        playcount = _get_track_playcount(track, args)

        if hasattr(args, "new_favorites_immediate_playcount_limit"):
            # Checking the attribute actually exists is required, since this is an optional parameter
            logging.debug("NewFavoritesImmediatePlaycountLimit defined")
            if playcount >= int(args.new_favorites_immediate_playcount_limit):
                old_favorites_tracks.append(track)

        if _has_track_reached_new_favorites_time_limit(track, args):
            if playcount <= int(args.new_favorites_playcount_limit):
                neglected_tracks.append(track)
            else:
                if track not in old_favorites_tracks:
                    old_favorites_tracks.append(track)

    logging.info(
        "Moving "
        + str(len(old_favorites_tracks))
        + " tracks from "
        + args.new_favorites_playlist
        + " to "
        + args.old_favorites_playlist
    )
    playlist.remove_tracks_from_playlist(
        spotify, args.new_favorites_playlist, old_favorites_tracks
    )
    playlist.add_tracks_to_playlist(
        spotify, args.old_favorites_playlist, old_favorites_tracks
    )

    logging.info(
        "Moving "
        + str(len(neglected_tracks))
        + " neglected tracks to "
        + args.neglected_playlist
    )
    playlist.remove_tracks_from_playlist(
        spotify, args.new_favorites_playlist, neglected_tracks
    )
    playlist.add_tracks_to_playlist(spotify, args.neglected_playlist, neglected_tracks)


def _get_track_playcount(track, args):
    try:
        return track_info.fetch_playcount(track, args.lastfm_user, args.lastfm_api_key)
    except:
        logging.warn("Couldn't get playcount for track " + str(track))
        return 0


def _has_track_reached_new_favorites_time_limit(track, args):
    added_at = iso8601.parse_date(track.added_at)
    track_age_in_playlist = added_at + relativedelta(
        days=int(args.new_favorites_time_limit)
    )
    return datetime.now(timezone.utc) >= track_age_in_playlist


def move_saved_tracks(spotify, args):
    saved_tracks = library.get_saved_tracks(spotify)
    tracks_to_move = []
    for track in saved_tracks:
        try:
            playcount = _get_track_playcount(track, args)
            if playcount >= int(args.saved_songs_playcount_limit):
                tracks_to_move.append(track)
        except:
            logging.warn("Couldn't get playcount for track " + str(track))

    logging.info(
        "Moving "
        + str(len(tracks_to_move))
        + " tracks from library to "
        + args.new_favorites_playlist
    )
    library.remove_tracks_from_library(spotify, tracks_to_move)
    playlist.add_tracks_to_playlist(
        spotify, args.new_favorites_playlist, tracks_to_move
    )


def _extract_args():
    args_parser = _setup_args_parser()
    args = args_parser.parse_args()
    args = _extract_user_configs(args)
    args = _extract_api_keys(args)
    return args


def _setup_args_parser():
    parser = argparse.ArgumentParser(
        description="Organize Spotify saved tracks by removing tracks with a certain playcount from liked tracks and moving them to other playlists"
    )
    parser.add_argument(
        "user_configs_file", type=argparse.FileType("r", encoding="UTF-8")
    )
    parser.add_argument("api_keys_file", type=argparse.FileType("r", encoding="UTF-8"))
    return parser


def _extract_user_configs(args):
    config_parser = ConfigParser()
    config_parser.read(args.user_configs_file.name)
    section = "Config"
    args.lastfm_user = config_parser[section]["LastFMUser"]
    args.spotify_user = config_parser[section]["SpotifyUser"]
    args.new_favorites_playlist = config_parser[section]["NewFavoritesPlaylist"]
    args.old_favorites_playlist = config_parser[section]["OldFavoritesPlaylist"]
    args.neglected_playlist = config_parser[section]["NeglectedPlaylist"]
    args.saved_songs_playcount_limit = config_parser[section][
        "SavedSongsPlaycountLimit"
    ]
    args.new_favorites_time_limit = config_parser[section]["NewFavoritesTimeLimit"]
    args.new_favorites_playcount_limit = config_parser[section][
        "NewFavoritesPlaycountLimit"
    ]

    try:
        args.new_favorites_immediate_playcount_limit = config_parser[section][
            "NewFavoritesImmediatePlaycountLimit"
        ]
    except KeyError:
        # Since this parameter is optional, we can just ignore KeyErrors thrown when it's not present
        logging.debug(
            "NewFavoritesImmediatePlaycountLimit not defined, so continuing without it"
        )

    return args


def _str_to_bool(to_convert):
    return to_convert == "True"


def _extract_api_keys(args):
    keys_parser = ApiKeysParser(args.api_keys_file)
    args.lastfm_api_key = keys_parser.lastfm_api_key
    args.spotify_client_id_key = keys_parser.spotify_client_id_key
    args.spotify_client_secret_key = keys_parser.spotify_client_secret_key
    return args


if __name__ == "__main__":
    organize_favorites()
